/*****************************************************************************
   mmnullaudio is a testing stub that implements an audio device
   that only writes messages to the error log (/var/log/messages). It
   is useful for diagnosing what OSS system calls an application makes.

   (c) 2006-2007 by Michael Minn - see www.michaelminn.com for contact info

   This program is free software; you can redistribute it and/or modify
   it under the terms of version 2 of the GNU General Public License as 
   published by the Free Software Foundation.

****************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/usb.h>
#include <linux/slab.h>
#include <linux/sound.h>
#include <asm/uaccess.h>
#include <linux/soundcard.h>
#include <linux/vermagic.h>
MODULE_INFO(vermagic, VERMAGIC_STRING);

static int mmnullaudio_audio_release (struct inode *inode, struct file *file)
{
	info("audio close()");
	return 0;
}

static int mmnullaudio_audio_open (struct inode *inode, struct file *file)
{
	// int device_number = MINOR(inode->i_rdev) >> 4;
	info("audio open (device 0x%x)", MINOR(inode->i_rdev));
	return 0;
}

static ssize_t mmnullaudio_audio_write (struct file *file, const char *data, size_t data_size, loff_t *offset)
{
	info("audio write(%d)", data_size);
	return data_size;
}

static ssize_t mmnullaudio_audio_read (struct file *file, char *data, size_t size, loff_t *offset)
{
	info("audio read %d bytes", size);
	return size;
}

static int mmnullaudio_audio_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	info("audio ioctl(0x%x)", command);
	return -EINVAL;
}

struct file_operations audio_fops =
{
	.read = mmnullaudio_audio_read, /* ssize_t (*read) (struct file *, char __user *, size_t, loff_t *); */
	.write = mmnullaudio_audio_write, /* ssize_t (*write) (struct file *, char __user *, size_t, loff_t *); */
	.ioctl = mmnullaudio_audio_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	.open = mmnullaudio_audio_open, /* int (*open) (struct inode *, struct file *); */
	.release = mmnullaudio_audio_release, /* int (*release) (struct inode *, struct file *); */
};


static int mmnullaudio_mixer_open (struct inode *inode, struct file *file)
{
	// int device_number = MINOR(inode->i_rdev) >> 4;
	info("mixer open (device 0x%x)", MINOR(inode->i_rdev));
	return 0;
}

static int mmnullaudio_mixer_release (struct inode *inode, struct file *file)
{
	info("mixer close()");
	return 0;
}

static ssize_t mmnullaudio_mixer_read(struct file *file, char __user *data, size_t size, loff_t *offset)
{
	info("mixer read()");
	return size;
}

static ssize_t mmnullaudio_mixer_write(struct file *file, const char *data, size_t size, loff_t *offset)
{
	info("mixer write()");
	return size;
}

static int mmnullaudio_mixer_ioctl(struct inode *node, struct file *file, unsigned int number, unsigned long data)
{
	info("mixer ioctl(0x%x)", number);
	return -EINVAL;
}

struct file_operations mixer_fops =
{
	.read = mmnullaudio_mixer_read, /* ssize_t (*read) (struct file *, char __user *, size_t, loff_t *); */
	.write = mmnullaudio_mixer_write, /* ssize_t (*write) (struct file *, char __user *, size_t, loff_t *); */
	.ioctl = mmnullaudio_mixer_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	.open = mmnullaudio_mixer_open, /* int (*open) (struct inode *, struct file *); */
	.release = mmnullaudio_mixer_release /* int (*release) (struct inode *, struct file *); */
};


static int null_audio_number = -1;
static int null_mixer_number = -1;

static int __init mmnullaudio_init(void)
{
	null_audio_number = register_sound_dsp(&audio_fops, -1);

	null_mixer_number = register_sound_mixer(&mixer_fops, -1);

	info("mmnullaudio uses dsp %d and mixer %d", null_audio_number, null_mixer_number);

	return 0;
}

static void __exit mmnullaudio_cleanup(void)
{
	if (null_audio_number >= 0)
		unregister_sound_dsp(null_audio_number);

	if (null_mixer_number >= 0)
		unregister_sound_mixer(null_mixer_number);

	info("mmnullaudio deregistered and unloaded");
}

module_init(mmnullaudio_init);
module_exit(mmnullaudio_cleanup);

